/*
 * entropy measurements using zlib
 * 2002 (c) marius aamodt eriksen (marius@umich.edu)
 * see copyright notice at the end of the file
 */

#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <zlib.h>

#include "config.h"
#include "netics.h"

#define ZERR(z) zerr((z), __FILE__, __LINE__)
#define ZWARN(z) zwarn((z), __FILE__, __LINE__)
#define ZBUFSZ(x) (ceil(1.01 * (x)))

struct lzw {
	z_stream  *z;
	u_char    *buf;
	u_int32_t  buflen;
	int        precise;
};

void   zerr(z_stream *, char *, int);
void   zwarn(z_stream *, char *, int);
double result(struct lzw *);
long   makeprecise(struct lzw *);

void *
lzw_new(u_int32_t buflen, int options)
{
	struct lzw *ret;

	ret = (struct lzw *)malloc(sizeof(struct lzw));

	ret->z = (z_stream *)malloc(sizeof(z_stream));

	ret->z->zalloc = Z_NULL;
	ret->z->zfree = Z_NULL;
	ret->z->avail_in = 0;
	ret->z->next_in = NULL;	
	
	if (deflateInit(ret->z, Z_BEST_COMPRESSION) != Z_OK) 
		ZERR(ret->z);
		
	ret->buf = (u_char *)malloc(ZBUFSZ(buflen));
	ret->buflen = ZBUFSZ(buflen);

	if (options & OPT_PRECISE)
		ret->precise = 1;
	else
		ret->precise = 0;

	return ((void *)ret);
}

void
lzw_free(void *v)
{
	struct lzw *l = (struct lzw *)v;

	free(l->z);
	free(l);
}

u_int32_t
lzw_feed(void *v, u_char *buf, u_int32_t buflen)
{
	struct lzw *l = (struct lzw *)v;
	
	l->z->avail_in = buflen;
	l->z->next_in = buf;
	l->z->next_out = l->buf;
	l->z->avail_out = ZBUFSZ(l->buflen);
	
	if (deflate(l->z, Z_NO_FLUSH) != Z_OK) 
		ZERR(l->z);

	/* we can always return a result */
	return (1);
}

double
lzw_result_raw(void *v)
{
	struct lzw *l = (struct lzw *)v;
	
	return result(l);
}

double
lzw_result_percent(void *v)
{
	struct lzw *l = (struct lzw *)v;
	double val;

	val = result(l);

	return (1 / val);
}


double
result(struct lzw *l)
{
	long out_len;
	double val = 0.0;

	/* cache results, use total_out + total_in as generation number */

        if (l->precise)
                out_len = makeprecise(l);
        else
                out_len = l->z->total_out;
        
        if (out_len != 0 && l->z->total_in != 0)
                val = (1.0 * l->z->total_in) / (1.0 * out_len);
	
	return (val);
}

long
makeprecise(struct lzw *l)
{
	z_stream *z = l->z;
	z_stream zview;
	int zstatus;
	long ret;
	char *out;

 	out = (char *)malloc(ZBUFSZ(l->buflen));

	assert(out != NULL);
	
	if (deflateCopy(&zview, z) != Z_OK)
		ZERR(z);

	zview.avail_out = ZBUFSZ(l->buflen);
	zview.next_out = out;

	while ((zstatus = deflate(&zview, Z_FINISH)) == Z_OK) {
		zview.avail_out = ZBUFSZ(l->buflen);
		zview.next_out = out;
	}
	
	if (zstatus != Z_STREAM_END) {
		warnx("precise measurement error");
		return (0);
	}

	ret = zview.total_out;
	free(out);
	deflateEnd(&zview);	

	return (ret);
}

void
lzw_reset(void *v)
{
	struct lzw *l = (struct lzw *)v;

	if (deflateReset(l->z) != Z_OK)
		ZWARN(l->z);
}

void
zerr(z_stream *z, char *file, int line)
{
	if (z->msg != NULL)
		errx(1, "%s:%i compression error: %s", file, line, z->msg);
	else
		errx(1, "%s:%i compression error", file, line);
}

void
zwarn(z_stream *z, char *file, int line)
{
	if (z->msg != NULL)
		warnx("%s:%i compression warning: %s", file, line, z->msg);
	else
		warnx("%s:%i compression warning", file, line);
}

struct sc_ops lzw_ops = {
	name:               "lzw",
	sc_new:            &lzw_new,
	sc_free:           &lzw_free,
	sc_feed:           &lzw_feed,
	sc_reset:          &lzw_reset,
	sc_result_raw:     &lzw_result_raw,
	sc_result_percent: &lzw_result_percent,
	sc_result_display: NULL
};

/*
 * copyright 2002 marius aamodt eriksen <marius@umich.edu>
 * all rights reserved.
 *
 * redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. all advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      this product includes software developed by marius aamodt eriksen.
 * 4. the name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
