/*
 * this implementation of the uli maurer's universal statistical test
 * algorithm was adopted from
 *    http://www.geocities.com/SiliconValley/Code/4704/uliscanc.txt
 * 
 * 2002 (c) marius aamodt eriksen (marius@umich.edu)
 * see end of file for copyright
 */

/*
 * XXX: better handling when result is ready.
 */

#include <sys/types.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "netics.h"

#define L 8
#define V (1 << L)
#define Q (10 * V)
#define K (100 * Q)
    
#define MAURER_EXPECTED 7.1836656

struct maurer {
	u_int32_t i;
	u_int32_t j;
	double    sum;
	double    result;
	int       table[V];
};

void * 
maurer_new(u_int32_t buflen, int options)
{
	struct maurer *ret;

	ret = (struct maurer *)malloc(sizeof(struct maurer));
	assert(ret != NULL);

	ret->i = 0;
	ret->j = Q;
	ret->sum = 0.0;
	ret->result = -1;
	memset(ret->table, 0, sizeof(ret->table));
	
	return ((void *)ret);
}


void
maurer_free(void *v)
{
	struct maurer *m = (struct maurer *)v;

	free(m);
}

/* XXX this needs to be changed so that we don't loose bytes */
u_int32_t
maurer_feed(void *v, u_char *buf, u_int32_t buflen)
{
	struct maurer *m = (struct maurer *)v;
	u_int32_t k = 0, q;
	
	/* fill initial table */
	if (m->i < Q) {
		for (; k < buflen && m->i < Q; m->i++, k++)
			m->table[buf[k]] = m->i;
		
		/* return negative for more bytes needed */
		if (m->i < Q)
			return -(Q - m->i);
	}
	
	for (; m->j < Q + K && k < buflen; m->j++, k++) {
		q = m->j;
		if (m->table[buf[k]] > q)
			q += K;
		m->sum += log((double)(m->j - m->table[buf[k]]));
		m->table[buf[k]] = m->j;
	}

	/* return non-negative when we have excess bytes, and a result
	 * is available */
	if (m->j == Q + K) {
		m->j = Q;
		m->result = (m->sum / (double)(K)) / log(2.0);
		m->sum = 0.0;
		return (buflen - k);
	}
	
	/* return 0 if we can yield a result */
	return (0);
}

void
maurer_reset(void *v)
{
	struct maurer *m = (struct maurer *)v;

	memset(m->table, 0, sizeof(m->table));
	m->j = Q;
	m->i = 0;
	m->sum = 0.0;
	m->result = -1.0;
}

double
maurer_result_percent(void *v)
{
	struct maurer *m = (struct maurer *)v;

	return (m->result / MAURER_EXPECTED);
}

double
maurer_result_raw(void *v)
{
	struct maurer *m = (struct maurer *)v;

	return (m->result);
}

struct sc_ops maurer_ops = {
	name:               "maurer",
	sc_new:            &maurer_new,
	sc_free:           &maurer_free,
	sc_feed:           &maurer_feed,
	sc_reset:          &maurer_reset,
	sc_result_raw:     &maurer_result_raw,
	sc_result_percent: &maurer_result_percent,
	sc_result_display: NULL
};

/*
 * copyright 2002 marius aamodt eriksen <marius@umich.edu>
 * all rights reserved.
 *
 * redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. all advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      this product includes software developed by marius aamodt eriksen.
 * 4. the name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
