/*
 * netics - the network statistics gatherer 
 * copyright (c) 2002 marius aamodt eriksen (marius@umich.edu)
 * see copyright at the end of the file
 */

/*
 * XXX TODO: chunk up measurements (e.g. every 10 kb, start new) (this
 * is done in maurer's scan).  clean up inconsistencies like where
 * conf is passed, and where it isn't.. should be passed to everything
 * but the signal handlers.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <assert.h>
#include <err.h>
#include <event.h>
#include <libnet.h>
#include <math.h>
#include <pcap.h>
#include <unistd.h>
#include <signal.h>

#include "config.h"
#include "netics.h"
#include "pcapu.h"

#define SNAPLEN 1500
#define C_TCP_H(x) ((struct libnet_tcp_hdr *)(x))
#define C_UDP_H(x) ((struct libnet_udp_hdr *)(x))
#define C_IP_H(x) ((struct libnet_ip_hdr *)(x))

#define MAJORVERSION 2
#define MINORVERSION 0

struct {
	int            lastint;
	int            numint;
	int            timeout;
	int            progressbar;
	struct sc_ops *ops;
	void          *handle;
} conf;

struct event sniffev, timeoutev;
extern struct sc_ops maurer_ops, lzw_ops;
struct sc_ops *scans[] = { &lzw_ops, &maurer_ops, NULL };

void usage(void);
void measure(u_char *);
void checkpoint(void);
void sig_cb(int);
int  evsig_cb(void);
void timeout_cb(int, short, void*);
void sniff_cb(int, short, void *);
void updatebar(double);

int
main(int argc, char **argv)
{
	int opt, sc_opt = 0, verbose = 0;
	char *intf = NULL, *file = NULL, *filter = "tcp or udp", *scan = "lzw";
	pcap_t *pcap_pd;
	extern char *__progname;
	struct timeval tv, timeout_tv;
	struct sc_ops **scani;
	extern int (*event_sigcb)(void);

	conf.timeout = 10;

	conf.progressbar = conf.lastint = conf.numint = 0;
 	conf.ops = NULL;

	event_init();

	while ((opt = getopt(argc, argv, "phVbvi:r:t:s:")) != -1)
 		switch (opt) {
		case 'i':
			intf = strdup(optarg);
			break;
		case 'p':
			sc_opt |= OPT_PRECISE;
			break;
		case 'r':
			file = strdup(optarg);
			break;
		case 't':
			conf.timeout = atoi(optarg);
			break;
		case 'b':
			conf.progressbar = 1;
			break;
		case 's':
			scan = strdup(optarg);
			break;
		case 'v':
			verbose++;
			break;
		case 'V':
			fprintf(stderr, "%s version %i.%i (%s)\n", __progname,
			    MAJORVERSION, MINORVERSION,
			    "$Id$");
			exit(0);
		case 'h':
		default:
			usage();
			/* NOTREACHED */
		}

	argc -= optind;
	argv += optind;
	
	scani = scans;

	while (*scani != NULL) {
		if (!strcmp((*scani)->name, scan)) {
			conf.ops = *scani;
			break;
		}
		scani++;
	}

	if (*scani == NULL)
		errx(1, "could not find scan %s", scan);

	if (verbose > 0)
		warnx("using %s scan", conf.ops->name);

	if (argc > 0)
		filter = copy_argv(argv);
	
	if ((pcap_pd = pcap_init(intf, file, filter, SNAPLEN)) == NULL)
		errx(1, "couldn't initialize sniffing");
	
	signal(SIGINT, &sig_cb);
	
	event_sigcb = &evsig_cb;
	
	conf.handle = (*conf.ops->sc_new)(SNAPLEN, sc_opt);

	timeout_set(&sniffev, sniff_cb, pcap_pd);
	timerclear(&tv);
	timeout_add(&sniffev, &tv);

	if (conf.timeout > 0) {
		timeout_set(&timeoutev, timeout_cb, NULL);
		timerclear(&timeout_tv);
		timeout_add(&timeoutev, &timeout_tv);
	} else
		errx(1, "invalid timeout value");

	event_dispatch();

	/* NOTREACHED */
	return (0);
}

void
sniff_cb(int fd, short event, void *arg)
{
	struct timeval tv;
	struct pcap_pkthdr h;
	u_char *p;
	int pcap_off;
	pcap_t *pcap_pd = (pcap_t *)arg;

	timerclear(&tv);
	
	if ((p = (u_char *)pcap_next(pcap_pd, &h)) != NULL) {
		if ((pcap_off = pcap_dloff(pcap_pd)) < 0)
			errx(1, "couldn't determine link layer offset");
		measure(p + pcap_off);
	} else
		tv.tv_usec = 1000 * 10;   /* 10 ms */
	
	timeout_add(&sniffev, &tv);
}

void
measure(u_char *p)
{
	struct libnet_ip_hdr *ip_h;
	u_char *data;
	int datalen;
	long in_len, prev_len;
	
	in_len = prev_len = 0;
	
	ip_h = (struct libnet_ip_hdr *)p;

	data = p + (ip_h->ip_hl << 2);

	if (ip_h->ip_p == IPPROTO_TCP)
		data += C_TCP_H(data)->th_off << 2;
	else if (ip_h->ip_p == IPPROTO_UDP) 
		data += LIBNET_UDP_H;
	else
		return;
	
	datalen = ntohs(ip_h->ip_len) - (data - p);
			
	if (datalen <= 0)
		return; 
			
	(*conf.ops->sc_feed)(conf.handle, data, datalen);
}

void
checkpoint(void)
{
	double val;

	if (conf.ops->sc_result_display != NULL) {
		(conf.ops->sc_result_display)(conf.handle);
		return;
	}

	if (conf.progressbar) {
		if ((val = (conf.ops->sc_result_percent)(conf.handle)) < 0)
			goto nodata;
		updatebar(val);
	} else {
		if ((val = (conf.ops->sc_result_raw)(conf.handle)) < 0) 
			goto nodata;
		printf("%d -> %f\n", time(NULL), val);
	}

	return;

 nodata:
	printf("\rdata not yet available");
	fflush(stdout);
}

int
evsig_cb(void)
{
	long curtime;
	
	if (conf.progressbar)
		exit(0);

	checkpoint();

	curtime = (int)time(NULL);

	if (curtime - conf.lastint < 2)
		conf.numint++;
	else
		conf.numint = 0;

	if (conf.numint == 2)
		exit(0);

	conf.lastint = curtime;

	return (1);
}

/* XXX make conf arg., pass it through.  only signal handlers should
 * use global conf */
void
timeout_cb(int fd, short event, void *arg)
{
	struct timeval tv;

	checkpoint();

	timerclear(&tv);
	tv.tv_sec = conf.timeout;
	event_add(&timeoutev, &tv);
}

void
sig_cb(int sig)
{
	extern int event_gotsig;
	
	event_gotsig = 1;
}

#define BUFSZ 256
void
updatebar(double value)
{
	u_int32_t width, progress, i;
	struct winsize winsz;
	char buf[BUFSZ];
	static int barpos = 0;
	static char *bars = "|/-\\";

	if (value < 0) {
		printf("\rdata not yet available");
		fflush(stdout);
		return;
	}
		
	if (ioctl(fileno(stdout), TIOCGWINSZ, &winsz) != -1)
		width = winsz.ws_col && winsz.ws_col < 256 ? winsz.ws_col : 80;
	else
		width = 80;
	
	width -= 10;
	
	if (value < 0) 
		value = 0.0;
	else if (value > 1.0) 
		value = 1.0;
	
	progress = ceil(value * width);
	
	snprintf(buf, sizeof(buf), "\r%3d%% |", (int)ceil(100 * value));
	
	i = strlen(buf);
	if (i + width > BUFSZ)
		return;
	memset(buf + i, '*', progress); i += progress;
	memset(buf + i, ' ', width - progress); i += width - progress;
	buf[i] = '\0';

	/* XXX do integers overflow on a multiple of 4? :) */
	snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), " |%c",
	    bars[barpos++ % sizeof(bars)]);
	
	/* XXX should be writen() -- import mio.c */
	write(fileno(stdout), buf, strlen(buf));
}
#undef BUFSZ /* 256 */

void
usage(void)
{
	extern char *__progname;
	struct sc_ops **scani;

	fprintf(stderr,
	    "%s: [-phVbv] [-i <interface>] [-r <file>] [-s <scan>] "
	    "[-t <timeout>] [filter]\n"
	    "\t-i <interface> sniff on specified interface\n"
	    "\t-r <file>      use tcpdump file <file> for sniffing\n"
	    "\t-t <timeout>   checkpoint every <timeout> seconds\n"
	    "\t-s <scan>      use scan <scan>.  available scans are:\n",
	    __progname);
	scani = scans;
	fprintf(stderr,
	    "\t               %s (default)", (*scani++)->name);
	while (*scani != NULL) {
		fprintf(stderr, ", %s", (*scani)->name);
		scani++;
	}
	fprintf(stderr, 	    
	    "\n"
	    "\t-p             use more precise measurements\n"
	    "\t-h             help\n"
	    "\t-V             print version number\n"
	    "\t-b             show measurement using a \"progress\" bar\n"
	    "\t-v             increase verbosity level\n");


	exit(1);
}

/*
 * copyright 2002 marius aamodt eriksen <marius@umich.edu>
 * all rights reserved.
 *
 * redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. all advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      this product includes software developed by marius aamodt eriksen.
 * 4. the name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
